Pro mage_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,combflatvar,$
                   dome=dome,diffuse=diffuse,bkspace=bkspace,arc=arc,$
                   refarc=refarc,offset=offset,scattered=scattered
                   
;+-------------------------------------------------------------------
;
; MAGE_FLATFIXER    2/2007
;
; Median-combine and normalize MagE flats.
;
; Input flats are median combined, the combined flat is normalized,
; and the formal variance in the normalized flat is computed assuming
; Poisson statistics.
;
; Assumes last 128 rows and columns are overscan (should be OK for
; any binning).
;
; Break point spacing for b-spline is set depending on the type of
; flat.  Xe flashlamps (the default) are fit with a narrow break point
; spacing to fit out spectral line.  Dome flats are fit with a wider
; spacing in order not to remove fringing features in the red.  Diffuse
; (out of focus) Xe flash lamp flats are fit using an intermediate spacing.
;       Type                           Spacing (pixels)
;       -----------------------------------------------
;       Internal Xe Flash (default)    1
;       Dome                           100
;       Diffuse Xe Flash               10    <---- Out of focus
;
; If you are fitting flats that use a different slit than you used
; for your science epxposures, you may want to set the offset keyword
; (see below).
;
; Errors in the flat field may be ~1 to a few % on the scales of the
; break point spacing, but should be near Poisson limit for pixel-to-
; pixel variations.
;
; NOTE: It looks like there are sharp spectral features in the dome 
;       flats at ~9500 ang that may be difficult to fit out without
;       also fitting out the fringes.
;
; INPUTS
;       flatlist  - String, name of single-column ASCII file containing
;                   the FITS filenames of flats to be combined, 
;                   OR string array of FITS filenames of flats to be
;                   combined.
;       wavearr   - 2D array or fits filename, where the pixel values
;                   are uncalibrated wavelengths (in pixels).
;       slitarr   - 2D array or fits filename, where the pixel values
;                   are position along the slit.
;       orderlist - Name of ASCII file containing the order
;                   information [CCD, order #, center, left edge,
;                   right edge], as output by mage_slitgrid.pro.
;
; KEYWORDS
;       dome      - If set, assumes these are dome flats.  A wide
;                   break point spacing is used.  Default is to assume
;                   that these are internal flats using the Xe flash
;                   lamp.
;       diffuse   - If set, assumes that these are out of focus flats
;                   (e.g., using the 5" slit).  Uses a wider break
;                   point spacing, and...  Default is to assume that
;                   these are in-focus flats.
;       bkspace   - Spacing of bspline break points, in pixels (or
;                   whatever units are used for the input wavearr).
;                   Defaults are specified above for internal and dome
;                   flats.  Smaller spacing may be needed for
;                   exposures that contain emission lines, but larger
;                   spacing should be used to preserve fringe
;                   features.
;      arc        - Arc exposure taken contemporaneously with flats
;      refarc     - Arc exposure taken with exposures used to create
;                   wavearr and slitarr.  If both arc and refarc
;                   keywords are fit, wavearr and slitarr will be registered
;                   to the flatfield exposures prior to fitting.
;      offset     - Scalar, an additional number of rows by which to shift
;                   the input wavearr and slitarr to match the setup
;                   of the flats.  This may be needed if you are fitting
;                   flats taken with a different slit than the one
;                   with which you took the arc (e.g., using a 5" slit
;                   to get more counts in the blue).  If the orders in the
;                   flats you are fitting appear to be shifted up (+y) 
;                   with respect to the traces followed by the arc, then
;                   the offset should be positive.  This may also be useful
;                   if you do not have an arc for your flats, and you
;                   need to make a rough offset by hand.
;       scattered - Subtract scattered light by fitting a 2D polynomial
;                   to the inter-order counts.
;
;  OUTPUTS
;       combflat    - Median-combined, normalized flat field.  May
;                     either be a variable or a fits filename to which
;                     array is to be written.
;       combflatvar - Formal variance in the combined, normlized flat.
;                     May either be a variable or a fits filename to
;                     which array is to be written.
;
; HISTORY
;       Adapted from MIKE_FLATFIXER 1/30/2008 GDB
;       SCATTERED keyword added 7/23/2009 GDB
;---------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:  '
   print,' mage_flatfixer,flatlist,wavearr,slitarr,orderlist,combflat,'
   print,'                combflatvar,/dome,/diffuse,bkspace=bkspace,'
   print,'                arc=arc,refarc=refarc,offset=offset,/scattered'
   return
endif

;;; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo       = read_ascii(orderlist,template=order_templ)
tmp_ordernum    = orderinfo.ordernum
tmp_ordercenter = orderinfo.center
tmp_orderedge1  = orderinfo.left
tmp_orderedge2  = orderinfo.right
; Sort orders in order of increasing slit position
order_order = sort(tmp_ordercenter)
ordernum    = tmp_ordernum(order_order)
ordercenter = tmp_ordercenter(order_order)
orderedge1  = tmp_orderedge1(order_order)
orderedge2  = tmp_orderedge2(order_order)
n_orders    = n_elements(ordercenter)

;;; Read in flat filenames
flatlist_sz = size(flatlist)
if (flatlist_sz(0) eq 0) then begin
   ; ASCII file with names of flat fields
   flat_templ = {   $
          version: 1.0, $
          datastart: 0L, $
          delimiter: ' ', $
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 1L, $
          fieldtypes: [7L], $
          fieldnames: ['name'], $
          fieldlocations: [6L], $
          fieldgroups: [0L] $
        }
   list     = read_ascii(flatlist,template=flat_templ)
   flatfile = list.name
endif else begin
   ; Array with names of flat fields
   flatfile = flatlist
endelse
nflat    = n_elements(flatfile)

;;; Read in wave and slit grids.

if (size(wavearr,/tname) eq 'STRING') then begin
   checkexists = findfile(wavearr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Wavelength file not found.'
      return
   endif else begin
      wavegrid = readfits(wavearr)
   endelse
endif else wavegrid = wavearr

if (size(slitarr,/tname) eq 'STRING') then begin
   checkexists = findfile(slitarr,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Slit position file not found.'
      return
   endif else begin
      slitgrid = readfits(slitarr)
   endelse
endif else slitgrid = slitarr

; Get size of arrays
sz  = size(wavegrid)
xsz = sz(1)
ysz = sz(2)

;;; Optional - Register wave and slit grids to flat field exposures

if (keyword_set(arc) and keyword_set(refarc)) then begin
   print,'Registering slit and wave grids to '+arc
   flat_arc = mage_subbias(readfits(arc))
   grid_arc = mage_subbias(readfits(refarc))
   mage_register,grid_arc,flat_arc,p,q
   print,'POLY_2D terms:'
   print,p,q
   ; Suppress pixels that were interpolated from order edges
   ordergrid = 1 + 0.*slitgrid
   interorder_ls = where(slitgrid eq -1000)
   ordergrid(interorder_ls) = 0.0   
   newordergrid = poly_2d(ordergrid,p,q,1,missing=0.0)
   slitgrid     = poly_2d(slitgrid,p,q,1,missing=-1000)
   wavegrid     = poly_2d(wavegrid,p,q,1,missing=-1000)
   suppress_ls  = where(newordergrid lt 1,n_suppress)
   if (n_suppress ne 0) then begin
      slitgrid(suppress_ls) = -1000
      wavegrid(suppress_ls) = -1000
   endif
endif

;;; Optional - Shift the wave and slit grids up or down by and
;;; additional amount, e.g., to match them to a flat that is offset
;;; from the input arc.

if keyword_set(offset) then begin
   offset        = round(offset)
   temp_slitgrid = shift(slitgrid,0,offset)
   temp_wavegrid = shift(wavegrid,0,offset)
   ; Fill in the missing rows by interpolating from the 
   ; nearest edge
   if (offset gt 0) then begin
      slit_deriv = slitgrid(*,0) - slitgrid(*,1)
      wave_deriv = wavegrid(*,0) - wavegrid(*,1)
      for i=0,offset-1 do begin
         temp_slitgrid(*,offset-1-i) = temp_slitgrid(*,offset-i) + slit_deriv
         temp_wavegrid(*,offset-1-i) = temp_wavegrid(*,offset-i) + wave_deriv
      endfor
   endif
   if (offset lt 0) then begin
      slit_deriv = slitgrid(*,ysz-1) - slitgrid(*,ysz-2)
      wave_deriv = wavegrid(*,ysz-1) - wavegrid(*,ysz-2)
      for i=0,-offset-1 do begin
         temp_slitgrid(*,ysz+offset+i) = temp_slitgrid(*,ysz+offset+i-1) + $
                                            slit_deriv
         temp_wavegrid(*,ysz+offset+i) = temp_wavegrid(*,ysz+offset+i-1) + $
                                            wave_deriv
      endfor
   endif
   slitgrid = temp_slitgrid
   wavegrid = temp_wavegrid
endif

;;;
;;; Median combine flats
;;;

; Arrays to store processed flats.
bigflat    = fltarr(xsz,ysz,nflat)
texp       = fltarr(nflat)         ; Exposure times

; Read in flats
for k=0,nflat-1 do begin
   tmp_flatfile = flatfile(k)
   checkexists  = findfile(tmp_flatfile,count=nfiles)
   if (nfiles eq 0) then begin
      print,'Flat file '+tmp_flatfile+' not found.'
      return
   endif else begin
      tmp_flat       = readfits(tmp_flatfile,flat_hdr)
      ; Subtract bias
      bigflat(*,*,k) = mage_subbias(tmp_flat)
      texp(k)        = sxpar(flat_hdr,'EXPTIME')
   endelse
endfor

; Determine gain and read noise
read_speed = sxpar(flat_hdr,'SPEED')
case read_speed of
   'Slow' : begin
               gn  = 1.02  ; e-/DN
               rdn = 2.9   ; e-
            end
   'Fast' : begin
               gn  = 0.82
               rdn = 3.1
            end
   'Turbo': begin
               gn  = 1.53
               rdn = 4.2
            end
   else:    begin
               print,'Unknown readout speed.  Assuming Fast.'
               gn  = 0.82
               rdn = 3.1
            end
endcase               

; Check that all flats have similar exposure times
texp_fracdiff = abs(texp - mean(texp))/mean(texp)
if (max(texp_fracdiff) gt 0.01) then begin
   print,'MAGE_FLATFIXER: Different exposure times detected - sky flats?'
   print,'                Scaling counts to match.'
   print,'                Note: Formal variance in combined flat may not'
   print,'                be correct.'
   temp_medflat = median(bigflat,dim=3,/even)
   ls           = where(temp_medflat ge 300)
   med_level    = median(temp_medflat(ls))
   for i=0,nflat-1 do $
      bigflat(*,*,i) = bigflat(*,*,i) * med_level / median(bigflat(*,*,i))
endif

; Median combine
if (nflat gt 1) then begin
   medflat = median(bigflat,dim=3,/even)
endif else begin
   medflat = bigflat
endelse

; Optional - Subtract scattered light
if keyword_set(scattered) then begin
   print,'Subtracting scattered light...'
   ; Select regions in between orders
   maskval    = -100
   interorder = maskval + 0.*medflat
   for i=0,n_orders-2 do begin
      inter_ls = where(slitgrid ge orderedge2(i)+5   and $
                       slitgrid le orderedge1(i+1)-5,n_inter)
      if (n_inter ne 0) then interorder(inter_ls) = medflat(inter_ls)
   endfor
   ; Include region below lowest order
   bottom_ls = where(slitgrid le orderedge1(0)-5,n_bottom)
   if (n_bottom ne 0) then interorder(bottom_ls) = medflat(bottom_ls)
   ; Mask non-illuminated regions
   size = size(slitgrid)
   xsz  = sz(1)
   ysz  = sz(2)
   xarr = fltarr(xsz,ysz)
   yarr = fltarr(xsz,ysz)
   for i=0,xsz-1 do xarr(i,*) = i
   for i=0,ysz-1 do yarr(*,i) = i
   nonillum_ls = where(xarr ge 2048 or yarr ge 1024,n_nonillum)  ; 1x1 binning
   if (n_nonillum ne 0) then interorder(nonillum_ls) = maskval
   ; Mask saturated regions.  Choose a low count threshhold, since
   ; scattered light should be relatively faint.
   saturated_ls = where(medflat ge 40000,n_saturated)
   if (n_saturated ne 0) then interorder(saturated_ls) = maskval
   ; Fit 2D polynomial
   fit_ls = where(interorder ne maskval,n_fit)
   if (n_fit eq 0) then begin
      print,'MAGE_FLATFIXER: No good pixels for fitting scattered light.'
      return
   endif
   interorder_data = fltarr(3,n_fit)
   interorder_data(0:1,*) = array_indices(interorder,fit_ls)
   interorder_data(2,*)   = interorder(fit_ls)
   fit_order = 4
   temp_fit  = sfit(interorder_data,fit_order,/irregular,kx=fitterms)
   ; Evaluate polynomial over entire array and subtract
   scattered_light = 0.*medflat
   for i=0,fit_order do begin
      for j=0,fit_order do begin
         scattered_light = scattered_light + fitterms(j,i)*(xarr^i)*(yarr^j)
      endfor
   endfor
   if (n_nonillum ne 0) then scattered_light(nonillum_ls) = 0.
   medflat = medflat - scattered_light
endif

;;;
;;; Normalize
;;;

; Set break point spacing for b-spline fit.
if keyword_set(bkspace) then bk_space = bkspace else begin
   bk_space = 1.   
   if keyword_set(diffuse) then bk_space = 10.
   if keyword_set(dome)    then bk_space = 100.
endelse

; Arrays of x and y-values
xarr = lonarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
yarr = lonarr(xsz,ysz)
for i=0,ysz-1 do yarr(*,i) = i
illum_xsz = xsz - 128
illum_ysz = ysz - 128

; Arrays of bad pixels/rows/columns.  Flag with 1 (none so far).
badarr = lonarr(xsz,ysz)

; Output arrays
normflat     =  1 + fltarr(xsz,ysz)
normflat_var = -1 + fltarr(xsz,ysz)

; Array of edges, useful for non-diffused flats.  For faint orders
; pixels in the center of the array can look like edges using the sobel
; method.  Therefore, construct a median "edge value" as a function
; of slit position for selecting pixels away from the edges.
edge_array = sobel(medflat)/float(medflat)
h          = histogram(slitgrid,bin=0.5,locations=slit_bin,rev=ri)
slit_bin   = slit_bin + 0.25
edge_bin   = 0.*slit_bin
for i=0,n_elements(h)-2 do if (h(i) ne 0) then $
   edge_bin(i) = median(edge_array(ri(ri(i):ri(i+1)-1)))

; Meausre a "minimum" variance from a small region between the first
; two orders
min_sdev = robust_sigma(medflat(550:650,100:110))
min_var  = min_sdev^2

;;; For each order, identify illuminated region and fit.

for i=0,n_orders-1 do begin

   print,' '
   print,'Order '+strtrim(ordernum(i),2)

   ; Identify illumated region
   slit_lo = orderedge1(i)
   slit_hi = orderedge2(i)
   edge_lo = min(slit_bin(where(slit_bin ge slit_lo and edge_bin lt 1.0)))
   edge_hi = max(slit_bin(where(slit_bin le slit_hi and edge_bin lt 1.0)))
   illum_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                    xarr le illum_xsz-1 and $
                    yarr le illum_ysz-1 and yarr ge 1,n_illum)
   
   ; Identify region to be fit.  Avoid edge regions.  For out of
   ; focus flats, fit a narrow region around the order center.
   if keyword_set(diffuse) then begin
      fit_ls = where(slitgrid ge ordercenter(i)-4 and $
                     slitgrid le ordercenter(i)+4 and $
                     xarr le illum_xsz-1 and $
                     yarr le illum_ysz-1 and yarr ge 1 and $
                     badarr ne 1,n_fit)
   endif else begin
      fit_ls = where(slitgrid ge (slit_lo > edge_lo) and $
                     slitgrid lt (slit_hi < edge_hi) and $
                     xarr le illum_xsz-1 and $
                     yarr le illum_ysz-1 and yarr ge 1 and $
                     badarr ne 1,n_fit)
   endelse
   fitslit = slitgrid(fit_ls)
   fitwave = wavegrid(fit_ls)
   fitflat = medflat(fit_ls)
   ; Assume Poisson noise (If there are two flats then they will
   ; be averaged, otherwise they are median combined.)
   if (nflat gt 2) then med_factor = !pi/2. else med_factor = 1
   fitflat_var = med_factor*(fitflat/gn + (rdn/gn)^2)/nflat
   fitflat_var = fitflat_var > min_var
   fitflat_invvar = 1./fitflat_var
   
   ; Identify potential bad pixels before fitting.
   print,'Pre-filtering bad pixels...'
   maxwave = max(fitwave)
   minwave = min(fitwave)
   first_filter = 1
   filter_sz = 2*bk_space
   n_filters = ceil((maxwave-minwave)/filter_sz)
   ; Default is not to filter any points
   filter_ls = -1
   ; Speed things up by dividing arrays into regions that contain
   ; several filter bins.  ***Should rewrite this using HISTOGRAM***
   filters_per_region = 1.*round(sqrt(n_filters))
   num_regions        = ceil(1.*n_filters / filters_per_region)
   for r=0,num_regions-1 do begin
      j_lo = r*filters_per_region
      j_hi = ((r+1)*filters_per_region - 1) < (n_filters - 1)
      region_wavelo = minwave + j_lo*filter_sz
      region_wavehi = minwave + (j_hi+1)*filter_sz
      region = where(fitwave ge region_wavelo and $
                     fitwave lt region_wavehi,n_region)
      if (n_region gt 0) then begin
         region_wave = fitwave(region)
         region_flat = fitflat(region)
         for j=j_lo,j_hi do begin
            wavelo = minwave + j*filter_sz
            wavehi = minwave + (j+1)*filter_sz
            bin_ls = where(region_wave ge wavelo and region_wave le wavehi,$
                           n_bin)
            if (n_bin gt 1) then begin
               tmp_flat   = region_flat(bin_ls) 
               tmp_median = median(tmp_flat)
               tmp_sdev   = robust_sigma(tmp_flat)
               tmp_filter_ls = where(abs(tmp_flat-tmp_median) ge 5*tmp_sdev,$
                                     n_tmp_filter)

               if (n_tmp_filter gt 0) then begin
                  if first_filter then begin
                     filter_ls    = region(bin_ls(tmp_filter_ls))
                     first_filter = 0 
                  endif else begin
                     filter_ls = [filter_ls,region(bin_ls(tmp_filter_ls))]
                  endelse
               endif
            endif
         endfor
      endif
   endfor
   if (total(filter_ls) ne -1) then fitflat_invvar(filter_ls) = 0

   ; Perform b-spline fit in the dispersion direction with no
   ; illumination function correction.
   sset = bspline_iterfit(fitwave,fitflat,invvar=fitflat_invvar,$
                          bkspace=bk_space,yfit=fitflat_model,upper=5,lower=5)
   ; Evaluate bspline fit over illuminated region and divide out
   medflat_model = fltarr(xsz,ysz)
   medflat_model(illum_ls) = bspline_valu(wavegrid(illum_ls),sset)
   normflat(illum_ls) = medflat(illum_ls) / medflat_model(illum_ls)
   
   ; Estimate variance in the normalized flat.  Non-illuminated regions 
   ; are flagged with -1.
   normflat_var(illum_ls) = $
         (med_factor*(medflat_model(illum_ls)/gn + (rdn/gn)^2)/nflat) / $
         (double(medflat_model(illum_ls))^2)

   ; Fit illumination function
   illumfit_ls = where(slitgrid ge slit_lo and slitgrid lt slit_hi and $
                       xarr le illum_xsz-1 and yarr le illum_ysz-1 and $
                       abs(normflat-1) le 0.5,n_illum)

   illum_slit       = slitgrid(illumfit_ls)
   illum_flat       = normflat(illumfit_ls)
   illumflat_invvar = 1./normflat_var(illumfit_ls)
   illum_sset = bspline_iterfit(illum_slit,illum_flat,invvar=illumflat_invvar,$
                                bkspace=1.0,yfit=illum_fit,upper=5,lower=5)

   ; Evalutate illumination function over entire slit and divide out
   full_illum_fit   = bspline_valu(slitgrid(illum_ls),illum_sset)
   
   normflat(illum_ls)     = normflat(illum_ls) / full_illum_fit
   normflat_var(illum_ls) = normflat_var(illum_ls) / (full_illum_fit^2)

endfor

;;;
;;; Set output arrays or write fits files
;;;

hnormflat = flat_hdr
sxaddhist,'-------------------------------------------------------',hnormflat
sxaddhist,'Flatfield processing done using MAGE_FLATFIXER',hnormflat
sxaddhist,'   Flatfield frames combined:',hnormflat
for i=0,nflat-1 do $
   sxaddhist,'      '+flatfile(i),hnormflat
sxaddhist,'-------------------------------------------------------',hnormflat

sxaddpar,hnormflat,'BZERO',0.0

if (size(combflat,/tname) eq 'STRING') then begin
   print,'Writing '+combflat
   writefits,combflat,normflat,hnormflat
endif else combflat = normflat

if (size(combflatvar,/tname) eq 'STRING') then begin
   print,'Writing '+combflatvar
   writefits,combflatvar,normflat_var,hnormflat
endif else combflatvar = normflat_var

print,'Done.'

return
end
