Pro mage_reduce,object,wavefile,slitfile,orderlist,skyref=skyref,$
                skyframe=skyframe,flatfield=flatfield,flatvar=flatvar,$
                redflatfield=redflatfield,redflatvar=redflatvar,output=output,$
                findrange=findrange,order=order,fit=fit,$
                fiducial=fiducial,fid_offset=fid_offset,fid_fwhm=fid_fwhm,$
                scattered=scattered,notrace=notrace,_extra=ex

;+-------------------------------------------------------------------
;
; MAGE_REDUCE     2/2007
;
; Reduce a MAGE frame using mutliple calls to longslit_reduce.
;
; INPUTS
;         object      - Object FITS filename
;         wavefile    - FITS filename of calibrated wavelength map
;                       (see mage_wavegrid.pro and
;                       mage_wavecalib.pro).  These should be AIR
;                       wavelengths.  Output wavelengths will be
;                       vacuum heliocentric.
;                       *** Currently no corrections made using skylines. ***
;         slitfile    - FITS filename of slit position map (see
;                       mage_slitgrid.pro)
;         orderlist   - Name of 5-column ASCII file containing, for
;                       each order, the the CCD name (blue or red),
;                       the order number (which is arbitrary, but
;                       should be consecutive integers), and should
;                       not be 0), the minimum slit position covered
;                       by the order, and the maximum slit position
;                       covered by the order.  Format is: [CCD, order
;                       #, center, left edge, right edge].  Output by
;                       mage_slitgrid.pro.
;
; KEYWORDS
;         skyref      - String, FITS filename of reference sky exposure
;                       for which the slit positions and wavelengths
;                       have been calibrated.  Should be in a similar
;                       setup to the object being reduced (shifts ~< 10
;                       pixels).  Skylines in the reference frame
;                       will be matched to skylines in the frame
;                       being reduced in order to register the slit
;                       positions and wavelengths to the current setup.
;         skyframe    - String.  If set, the skylines in skyref will
;                       be matched to the skylines in this exposure,
;                       rather than to those in the frame being
;                       reduced.  This exposure should therefore be
;                       taken in exdactly the same setup as the
;                       exposure being reduced.  Useful for standard
;                       star exposures that may not have pronounced
;                       skylines.
;         flatfield   - FITS filename of flat field array
;         flatvar     - FITS filename of flat field variance array
;         redflatfield- FITS filename of flat field to be used for red
;                       orders (quartz dome flat)
;         redflatvar  - FITS filename of variance array for flat field
;                       to be used for red orders
;         output      - Root name for output files.  Default is
;                       'mage'.
;         findrange   - Two-elements array giving the min and max x-values
;                       over which to locate the object in the fiducial
;                       order.  DISABLED.
;         order       - 1D array specifying the orders to be reduced.
;                       Numbers correspond to the entry number in the
;                       input orderlist (1 being the first order listed
;                       in the file).  Must include the fiducial order.
;         fit         - 1D array specifying the orders in which the
;                       object should be indepently located and traced
;                       (unless the /notrace keyword is set), and its
;                       FWHM measured.  These orders should have
;                       significant flux over most of the length of
;                       the array.  Other orders will have their
;                       object position, trace, and FWHM extrapolated
;                       from the fitted orders, which is useful for
;                       orders with little or no flux.  Default is to
;                       independently fit all orders.  Must include
;                       the fiducial order.
;         fiducial    - Scalar, order number to be used as the
;                       fiducial order (the order number is determined
;                       by the order number entry in the input
;                       orderlist).  This order will be reduced first
;                       and willl be used to measure the initial
;                       profile FWHM (for Gaussian profiles).  The
;                       default is to use one of the middle orders.
;                       However, it is recommended to use the order
;                       that contains the most flux.
;         fid_offset  - Scalar, offset from the order center (as
;                       defined in the input orderlist), in pixels, of
;                       the object in the fiducial order.  Can be set
;                       along with fid_fwhm instead of attempting to
;                       automatically fit the object profile.
;         fid_fwhm    - Scalar, spatial Gaussian FWHM, in pixels,
;                       of the object in the fiducial order.  Can be set
;                       along with fid_offset instead of attempting to
;                       automatically fit the object profile.
;         scattered   - Subtract scattered light by fitting a 2D polynomial
;                       to the inter-order counts.
;         notrace     - If set, do not attempt to trace object.  Fits will
;                       be made to the slit position and FWHM only.
;         _extra      - Additional keywords accepted by
;                       longslit_reduce, e.g., sprofile, maxiter,
;                       noshow, and noskysub.  Note that some keywords
;                       may not work due to their use in the program
;                       body below.
;
; OUTPUTS
;         Outputs written to output_sub.fits, output_spec.fits, etc.
;
; Side effects: If there is a journal file currently open, the program
;               will close it and open <output>.log.
;
; HISTORY
;         Adapted from MIKE_REDUCE and NEW_HIRES_REDUCE 2/1/2008 GDB
;         ARC keyword added 5/14/2007 GDB
;         Added NOTRACE keyword, and adapted trace decision algorithm
;            from ESI_REDUCE 9/16/2011 GDB
;--------------------------------------------------------------------------

if (n_params() eq 0) then begin
   print,'CALLING SEQUENCE:'
   print,' mage_reduce,object,wavefile,slitfile,orderlist,skyref=skyref,'
   print,'        skyframe=skyframe,flatfield=flatfield,flatvar=flatvar,'
   print,'        redflatfield=redflatfield,redflatvar=redflatvar,'
   print,'        output=output,findrange=findrange,order=order,fit=fit,'
   print,'        fiducial=fiducial,fid_offset=fid_offset,fid_fwhm=fid_fwhm,'
   print,'        /scattered,/notrace,_extra=ex'
   return
endif

; Root name for output
if keyword_set(output) then root = output else root = 'mage'

; Start log.  If there is a journal already running, close it and start a
; new one.
if (!journal ne 0) then journal
journal,root+'.log'

print,''
print,'Reducing '+object
print,''

; Read in FITS files
obj  = readfits(object,hobj)
wave = readfits(wavefile,hwave)
slit = readfits(slitfile,hslit)

; Bias-subtract
obj = mage_subbias(obj)

; Get size of individual CCD arrays
sz   = size(obj)
xsz  = sz(1)
ysz  = sz(2)
xarr = fltarr(xsz,ysz)
yarr = fltarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i

; Determine gain and read noise
read_speed = sxpar(hobj,'SPEED')
case read_speed of
   'Slow' : begin
               gain    = 1.02  ; e-/DN
               rdnoise = 2.9   ; e-
            end
   'Fast' : begin
               gain    = 0.82
               rdnoise = 3.1
            end
   'Turbo': begin
               gain    = 1.53
               rdnoise = 4.2
            end
   else:    begin
               print,'Unknown readout speed.  Assuming Fast.'
               gain    = 0.82
               rdnoise = 3.1
            end
endcase               

;;; Register wavelength and slit position grids to current exposure
;;; using skylines.  Flag pixels that are not calibrated.
if not(keyword_set(skyref)) then begin
   print,'Skipping registering slit and wavelength arrays.'
endif else begin
   print,'Registering slit and wavelength arrays...'
   sky_ref = mage_subbias(readfits(skyref))
   if keyword_set(skyframe) then begin
      ; Optional - Loacte skylines in a separate exposure in the same
      ; setup
      print,'Registering to '+skyframe
      sky_frame = mage_subbias(readfits(skyframe))
      mage_register,sky_ref,sky_frame,p,q
   endif else begin
      mage_register,sky_ref,obj,p,q
   endelse
   print,'POLY_2D terms: '
   print,p,q
   ; Mask uncalibrated pixels with NaN values.  This will tend to lose
   ; a few pixels at the edges of the array.  It may be possible to 
   ; interpolate the wavelength solution instead, and fill in the missing
   ; values, but this would only be worthwhile if the shift was large
   ; (> 10 pixels in wavelength).
   ; Keep track of uncalibrated pixels
   calibgrid = 1 + 0.*slit
   notcalib_ls = where(slit lt -999 or wave lt 0,n_notcalib)
   if (n_notcalib ne 0) then calibgrid(notcalib_ls) = 0.0
   newcalibgrid = poly_2d(calibgrid,p,q,1,missing=0.0)
   nan   = !values.f_nan
   wave = poly_2d(wave,p,q,1,missing=nan)
   slit = poly_2d(slit,p,q,1,missing=nan)
   new_notcalib_ls = where(newcalibgrid lt 1,n_new_notcalib)
   if (n_new_notcalib ne 0) then begin
      wave(new_notcalib_ls) = nan
      slit(new_notcalib_ls) = nan
   endif
   ;; Suppress pixels that were interpolated from uncalibrated pixels,
   ;; which should have been flagged.
   ;wave_ls = where(wave lt 0,n_wave)
   ;if (n_wave ne 0) then wave(wave_ls) = nan
   ;slit_ls = where(slit lt 0,n_slit)
   ;if (n_slit ne 0) then slit(slit_ls) = nan
endelse

; Read in flat files.  Assume bias has already been subtracted.
if keyword_set(flatfield) then flat = readfits(flatfield,hflat) $
   else flat = 0
if keyword_set(flatvar) then flat_var = readfits(flatvar,hflatvar) $
   else flat_var = 0

; Read in flat field for red orders
if keyword_set(redflatfield) then begin
   red_flat = readfits(redflatfield,hredflat)
   red_ls   = where(slit ge 400)   ; This should be between orders
   if n_elements(flat eq 0) then flat = 1. + fltarr(xsz,ysz)
   flat(red_ls) = red_flat(red_ls)
endif
if keyword_set(redflatvar) then begin
   red_flat_var = readfits(redflatvar,hredflatvar)
   red_ls   = where(slit ge 400)   ; This should be between orders
   if n_elements(flat_var eq 0) then flat_var = 1e-6 + fltarr(xsz,ysz)
   flat_var(red_ls) = red_flat_var(red_ls)
endif

; Read in order information
order_templ =  {   $
          version: 1.0, $
          datastart: 0, $
          delimiter: ' ', $        ; Fields separated by white space
          missingvalue: -99.0, $
          commentsymbol: '#', $
          fieldcount: 5, $
          fieldtypes: [7,4,4,4,4], $
          fieldnames: ['ccd','ordernum','center','left','right'], $
          fieldlocations: [2,11,18,30,42], $
          fieldgroups: [0,1,2,3,4] $
        }
orderinfo       = read_ascii(orderlist,template=order_templ)
; Make sure all order numbers are unique
if (n_elements(orderinfo.ordernum) ne $
    n_elements(uniq(orderinfo.ordernum))) then begin
   print,'Error - All order numbers must be unique.'
   return
endif
; Sort orders in order of increasing slit position
sort_ls     = sort(orderinfo.center)
ccdname     = orderinfo.ccd(sort_ls)
ordernum    = orderinfo.ordernum(sort_ls)
ordercenter = orderinfo.center(sort_ls)
orderedge1  = orderinfo.left(sort_ls)
orderedge2  = orderinfo.right(sort_ls)
n_orders    = n_elements(ordercenter)

; Optional - Specify the orders to reduce.
if keyword_set(order) then begin
   n_reduce    = n_elements(order)
   reduce_ls   = intarr(n_reduce)
   for i=0,n_reduce-1 do reduce_ls(i) = where(ordernum eq order(i))
   ordercenter = ordercenter(reduce_ls)
   ccdname     = ccdname(reduce_ls)
   ordernum    = ordernum(reduce_ls)
   orderedge1  = orderedge1(reduce_ls)
   orderedge2  = orderedge2(reduce_ls)
   n_orders    = n_elements(ordercenter)
endif

; Optional - Specify orders the orders for which to independently
; locate and trace the object, and measure its FWHM.  Default is
; to independently fit all orders.
do_independent_fit = lonarr(n_orders)
if keyword_set(fit) then begin
   for i=0,n_elements(fit)-1 do begin
      j = where(ordernum eq fit(i))
      do_independent_fit(j) = 1
   endfor
endif else begin
   do_independent_fit(*) = 1
endelse

; Keep track of orders for which independent tracing and fitting
; has been done.
independent_fit_made = lonarr(n_orders)

; Optional - Specify the order to be reduced first.  Default is
; to start with an order in the middle.
if keyword_set(fiducial) then begin
   fid_index = where(ordernum eq fiducial,n_fid)
   fid_index = fid_index(0)
   if (n_fid eq 0) then begin
      print,'Error - Fiducial order must be included in the orders to reduce.'
      return
   endif
endif else fid_index = floor(n_orders/2.)

; Flag bad pixels to be masked by longslit_reduce.
; (None)

; Optional - Set x-range over which to locate object.
; Note: nominal_find_range not currently being used.
if keyword_set(findrange) then nominal_find_range=findrange else $
   nominal_find_range=[900,1100]

; Optional - Subtract scattered light
if keyword_set(scattered) then begin
   print,'Subtracting scattered light...'
   ; Select regions in between orders
   maskval    = -100
   interorder = maskval + 0.*obj
   for i=0,n_orders-2 do begin
      inter_ls = where(slit ge orderedge2(i)+5   and $
                       slit le orderedge1(i+1)-5,n_inter)
      if (n_inter ne 0) then interorder(inter_ls) = obj(inter_ls)
   endfor
   ; Include region below lowest order
   bottom_ls = where(slit le orderedge1(0)-5,n_bottom)
   if (n_bottom ne 0) then interorder(bottom_ls) = obj(bottom_ls)
   ; Mask non-illuminated regions
   nonillum_ls = where(xarr ge 2048 or yarr ge 1024,n_nonillum)  ; 1x1 binning
   if (n_nonillum ne 0) then interorder(nonillum_ls) = maskval
   ; Fit 2D polynomial
   fit_ls = where(interorder ne maskval,n_fit)
   if (n_fit eq 0) then begin
      print,'MAGE_REDUCE: No good pixels for fitting scattered light.'
      return
   endif
   interorder_data = fltarr(3,n_fit)
   interorder_data(0:1,*) = array_indices(interorder,fit_ls)
   interorder_data(2,*)   = interorder(fit_ls)
   fit_order = 4
   temp_fit  = sfit(interorder_data,fit_order,/irregular,kx=fitterms)
   ; Evaluate polynomial over entire array and subtract
   scattered_light = 0.*obj
   for i=0,fit_order do begin
      for j=0,fit_order do begin
         scattered_light = scattered_light + fitterms(j,i)*(xarr^i)*(yarr^j)
      endfor
   endfor
   if (n_nonillum ne 0) then scattered_light(nonillum_ls) = 0.
   obj = obj - scattered_light
endif

; Initalize large spectrum, error, and wavelength
; arrays in which to store the individual reduced orders.
big_wave = dblarr(10000,n_orders)
big_spec = dblarr(10000,n_orders)
big_err  = dblarr(10000,n_orders)

; Keep track of the median wavelength, Gaussian profile FWHM, slit
; position, offset from the slit center, and trace coefficients (if
; tracing object) for each order.
medwave_list   = dblarr(n_orders)
fwhm_list      = dblarr(n_orders)
position_list  = dblarr(n_orders)
offset_list    = dblarr(n_orders)
trace_coeffs   = dblarr(5,n_orders)

; Initialize fits to FWHM, position, offset from the slit center,
; and linear trace term
fwhm_fit     = dblarr(n_orders)
position_fit = dblarr(n_orders)
offset_fit   = dblarr(n_orders)
slope_fit    = dblarr(n_orders)

; Initialize reduced arrays
nosky    = fltarr(xsz,ysz)
noskyvar = fltarr(xsz,ysz)
wavefit  = dblarr(xsz,ysz)
skymodel = fltarr(xsz,ysz)

;; Set direction of increasing wavelength.  Start by stepping through
;; orders in the direction of increaseing wavelength.
;red_direction  =  1
;blue_direction = -1

; High order terms for trace trace parameters as a function of order
; number.  Use a rough linear term for FWHM.  This was measured from
; fits to a standard star, and should be at least be better than
; assuming the FWHM is constant with wavelenth.  For now, ignore
; higher order terms for the offset from slit center and the slope.
fwhm_hot     = 5.0 - 0.146*ordernum
offset_hot   = 0.0
slope_hot    = 0.0

; Loop through orders.  Start with fiducial order
index_list = [fid_index]

; Next do indepent orders redward of the fiducial order
red_indep_ls = where(ordernum gt fiducial and do_independent_fit,n_red_indep)
if (n_red_indep ne 0) then index_list = [index_list,red_indep_ls]

; Next do indepent orders blueward of the fiducial order, going
; red to blue
blue_indep_ls = where(ordernum lt fiducial and do_independent_fit,n_blue_indep)
if (n_blue_indep ne 0) then index_list = [index_list,reverse(blue_indep_ls)]

; Finally, do non-independent orders (order in which these are done
; does not matter)
non_indep_ls = where(do_independent_fit eq 0,n_non_indep)
if (n_non_indep ne 0) then index_list = [index_list,non_indep_ls]

for i=0,n_orders-1 do begin

   index = index_list(i)

   print,' '
   print,'Reducing '+object+', order '+strtrim(ordernum(index),2)
   print,' '

   ; Extrapolate fit and trace parameters from orders that have been
   ; independently fit.  If there are very few orders for which these
   ; quantities can be measured independently, do a constant offset
   ; only, possibly including pre-measured higher-order terms.  This
   ; is equivalent to doing an offset from the standard star from
   ; which the slit grid was created.  In that case, the extrapolated
   ; values may only be reasonable out to a few orders.  When many
   ; orders are fit independently, fit a linear or higher-order trend.
   ; Quadratic fits seem be good for FWHM and offset.  Slope has not
   ; yet been looked at carefully.  NOTE: The offset seems to require
   ; a lower-order polynomial fit than the raw position (which makes
   ; it easier and more reliable to fit).
   indep_ls     = where(independent_fit_made eq 1,n_indep)
   n_for_linear = 6
   n_for_quad   = 8
   if (n_indep ge n_for_linear) then begin
      if (n_indep ge n_for_quad) then begin
         fwhm_order     = 2
         offset_order   = 2      ; Fit offset from slit center rather than
         slope_order    = 2 ;1?  ; slit position itself.
      endif else begin
         fwhm_order     = 1
         offset_order   = 1
         slope_order    = 1
      endelse
   endif else begin
      fwhm_order     = 0
      offset_order   = 0
      slope_order    = 0
   endelse
   if (n_indep ge 1) then begin
      ; Subtract higher order terms (if any) from fit parameters.
      net_fwhm     = fwhm_list         - fwhm_hot
      net_offset   = offset_list       - offset_hot
      net_slope    = trace_coeffs(1,*) - slope_hot
      ; Fit residuals
      if (n_indep ge n_for_linear) then begin
         pfit_fwhm     = robust_poly_fit(ordernum(indep_ls),$
                                         net_fwhm(indep_ls),$
                                         fwhm_order)
         pfit_offset = robust_poly_fit(ordernum(indep_ls),$
                                         net_offset(indep_ls),$
                                         offset_order)
         pfit_slope    = robust_poly_fit(ordernum(indep_ls),$
                                         net_slope(indep_ls),$
                                         slope_order)
      endif else begin
         pfit_fwhm     = poly_fit(ordernum(indep_ls),net_fwhm(indep_ls),$
                                  fwhm_order)
         pfit_offset = poly_fit(ordernum(indep_ls),net_offset(indep_ls),$
                                  offset_order)
         pfit_slope    = poly_fit(ordernum(indep_ls),net_slope(indep_ls),$
                                         slope_order)  
      endelse
      ; Reconstruct total fitted values
      fwhm_fit     = fwhm_hot
      offset_fit   = offset_hot
      slope_fit    = slope_hot
      for k=0,fwhm_order do $
         fwhm_fit = fwhm_fit + pfit_fwhm(k)*(ordernum^k)
      for k=0,offset_order do $
         offset_fit = offset_fit + pfit_offset(k)*(ordernum^k)
      for k=0,slope_order do $
         slope_fit = slope_fit + pfit_slope(k)*(ordernum^k)
   endif else begin
      fwhm_fit     = 0.*ordernum
      offset_fit   = 0.*ordernum
      slope_fit    = 0.*ordernum
   endelse
   ; Compoute actual slit positions
   position_fit = ordercenter + offset_fit

   ; For the first order, start with no estimates for the object
   ; profile FWHM.  For subsequent orders that are being independently
   ; traced and fit, start with the fitted FWHM estimate.  For orders
   ; that are not to be independently fit, use extrapolated values for
   ; the object position, object FWHM, and trace slope from orders for
   ; which independent fits have been made.
   if (index eq fid_index) then begin
      slitpos = 0   ; Return fitted slit position
      fwhm    = 0   ; FWHM estimate
      slope   = 0
      force   = 0
      independent_fit_made(index) = 1
      ; Optional - User-specified values
      if (keyword_set(fid_offset) or keyword_set(fid_fwhm)) then begin
         if (keyword_set(fid_offset) and keyword_set(fid_fwhm)) then begin
            slitpos   = ordercenter(index) + fid_offset
            fwhm      = fid_fwhm
            force     = 1
         endif else begin
            print,'Must specify both fid_offset and fid_fwhm (or neither)'
            return
         endelse
      endif
   endif else begin
      if (do_independent_fit(index) eq 1) then begin
         slitpos = 0
         fwhm    = fwhm_fit(index)
         slope   = 0
         force   = 0
         independent_fit_made(index) = 1
      endif else begin
         ; Use fitted values
         slitpos = position_fit(index)
         fwhm    = fwhm_fit(index)
         slope   = slope_fit(index)
         force   = 1
         print,'Using fitted values: Position = '+strtrim(slitpos,2)
         print,'                     FWHM     = '+strtrim(fwhm,2)
         print,'                     Slope    = '+strtrim(slope,2)
         print,' '
      endelse
   endelse

   ;; If using an estimate for the slope of the object trace, adjust
   ;; slit grid before passing to longslit_reduce.  If not, trace
   ;; the object.
   ;temp_slit = slit
   ;if (slope ne 0) then begin
   ;   slit_correction = slope*findgen(xsz)
   ;   for j=0,xsz-1 do temp_slit(j,*) = slit(j,*) - slit_correction(j)
   ;   do_trace = 0
   ;endif else begin
   ;   do_trace = 1
   ;endelse
   ;
   ;print,'Not tracing orders where independent fits are not being made'
   ;print,'is handled better in esi_reduce.  Should do_trace be -1?'
   ;stop

   ; If using an estimate for the slope of the object trace, adjust
   ; slit grid before passing to longslit_reduce.  If not, trace
   ; the object.
   temp_slit = slit
   if (slope ne 0) then begin
      slit_correction = slope*findgen(ysz)
      for j=0,ysz-1 do temp_slit(*,j) = slit(*,j) - slit_correction(j)
   endif

   ; If using fitted values, make sure tracing is turned off in
   ; longslit_reduce.
   if force then begin
      do_trace = -1
   endif else begin
      ; Default is set by the value of traceobj in the
      ; longslit_reduce.inc file, unless the /notrace keyword is set.
      if keyword_set(notrace) then do_trace = -1 else do_trace = 0
   endelse

   ; Note: If the trace_obj keyword is set in the call to mage_reduce,
   ; then this will be passed to longslit_reduce if do_trace=0 (can
   ; be used to prevent longslit_reduce from tracing object). Better
   ; handled by setting /notrace keyword, however.

   ; Pass header to longslit_reduce
   temp_hdr = hobj

   ; Set approximate slit positions
   temp_ordercenter = ordercenter(index)
   temp_slitmin     = orderedge1(index)
   temp_slitmax     = orderedge2(index)

   ; Set break point spacing according to approximate dispersion
   this_order   = where(slit ge temp_slitmin and slit le temp_slitmax and $
                        wave gt 100,n_this_order)
   rough_disp   = (max(wave(this_order))-min(wave(this_order))) / $
                  (max(xarr(this_order))-min(xarr(this_order))) 
   bk_space     = 3*rough_disp
   min_bk_space = 0.7*rough_disp
   print,'Setting break point spacing = '+strtrim(bk_space,2)
   print,'    min break point spacing = '+strtrim(min_bk_space,2)

   ; If necessary, override region of the chip in which to locate
   ; object.  This is useful for tracing standards in non-fiducial
   ; orders.
   trace_range = [0,xsz-1-128]

   ; Do extraction in wavelength bins of constant velocity width
   log_lin = 1

   ; Reduce order.  longslit_reduce will convert input wavelengths to
   ; vacuum heliocentric.

   longslit_reduce,obj,wave,temp_slit,temp_nosky,temp_noskyvar,$
      temp_wavefit,temp_skymodel,hdr=temp_hdr,/mage,/nobias,$
      flatfield=flat,flatvar=flat_var,gn=gain,rdnoise=rdnoise,$
      fwhm=fwhm,slitpos=slitpos,force=force,trace_obj=do_trace,$
      meanshift=mean_shift,outfile='temp_mage',$
      tracerange=trace_range,subrange=[temp_slitmin,temp_slitmax],$
      findrange=find_range,waveref=wave,/vacuum,$
      bkspace=bk_space,min_bkspace=min_bk_space,/set_bkpts,loglin=log_lin,$
      _extra=ex

   ; Check - Did longslit_reduce run properly?
   if (n_elements(temp_nosky) eq 1 or $
       n_elements(temp_noskyvar) eq 1 or $
       n_elements(temp_wavefit) eq 1 or $
       n_elements(temp_skymodel) eq 1) $
      then longslit_ok = 0 else longslit_ok = 1

   if longslit_ok then begin

      ; Store the section of the 2D reduced frames for this order.
      order_ls = where(slit ge temp_slitmin and slit le temp_slitmax)
      nosky(order_ls)    = temp_nosky(order_ls)
      noskyvar(order_ls) = temp_noskyvar(order_ls)
      wavefit(order_ls)  = temp_wavefit(order_ls)
      skymodel(order_ls) = temp_skymodel(order_ls)
      
      ; Read the new spectrum back in and store in the large array.
      checkexists = findfile('temp_mage_spec.fits',count=n_specfiles)
      checkexists = findfile('temp_mage_err.fits',count=n_errfiles)
      if not(n_specfiles eq 0 or n_errfiles eq 0) then begin
         spec = rflam('temp_mage_spec.fits',hspec,wspec)
         err  = rflam('temp_mage_err.fits',herr,werr)
         n_pix = n_elements(spec)
         big_wave(0:n_pix-1,index) = wspec
         big_spec(0:n_pix-1,index) = spec
         big_err(0:n_pix-1,index)  = err
      endif
      
      ; Record the median wavelength, profile FWHM, slit position,
      ; offset from the standard star, and mean wavelength shift
      ; for this order.  Calculate offset from slit center.
      medwave_list(index)   = median(wspec,/even)
      fwhm_list(index)      = fwhm
      position_list(index)  = slitpos
      offset_list(index)    = slitpos - ordercenter(index)

      ; If a trace was made, read in and store trace information
      if do_trace then begin
         trace_templ = {   $
                   version: 1.0, $    
                   datastart: 0L, $   
                   delimiter: ' ', $
                   missingvalue: -99.0, $
                   commentsymbol: '#', $
                   fieldcount: 1L, $
                   fieldtypes: 4L, $
                   fieldnames: 'value', $
                   fieldlocations: 0L, $
                   fieldgroups: [0L] $
                 }
         trace_data = read_ascii('temp_mage_trace.dat',template=trace_templ)
         ; The first numbers are the polynomial fit to the object trace.  The
         ; last number is the object centroid, which should be nearly identical
         ; to the zeroeth term in the trace coefficients.
         all_numbers  = trace_data.value
         n_numbers    = n_elements(all_numbers)
         temp_trace_coeffs = all_numbers(0:n_numbers-2)
         n_coeffs     = n_elements(temp_trace_coeffs)
         trace_coeffs(0:n_coeffs-1,index) = temp_trace_coeffs
      endif else begin
         ; Record fitted values
         trace_coeffs(0:1,index) = [slitpos,slope]
      endelse

   endif

endfor

; Save the 2D reduced arrays.
hout = hobj
sxaddpar,hout,'BZERO',0.0
get_date,date,/timetag
sxaddhist,'Reduced with MAGE_REDUCE '+date+' (UTC)',hout
sxaddpar,hout,'ORIGEXP',filename(object),'Filename of original exposure.'
sxaddpar,hout,'WAVEGRID',filename(wavefile),'Wavelength-proxy grid'
sxaddpar,hout,'SLITGRID',filename(slitfile),'Slit-position-proxy grid'
if keyword_set(arc) then $
   sxaddpar,hout,'ARC',filename(arc),'Comparison lamp'
if keyword_set(flatfield) then $
   sxaddpar,hout,'FLATFLD',filename(flatfield),'Flat field'
if keyword_set(flatvar) then $
   sxaddpar,hout,'FLATVAR',filename(flatvar),'Flat field variance'
writefits,root+'_sub.fits',nosky,hout
writefits,root+'_var.fits',noskyvar,hout
writefits,root+'_sky.fits',skymodel,hout
writefits,root+'_wav.fits',wavefit,hout

; Remove extra zeros and save the big spec and error arrays.
if (n_orders gt 1) then begin
   nonzero_ls  = where(total(big_wave,2) gt 0)
   max_nonzero = max(nonzero_ls)
   big_wave    = big_wave(0:max_nonzero-1,*)
   big_spec    = float(big_spec(0:max_nonzero-1,*))
   big_err     = float(big_err(0:max_nonzero-1,*))
endif else begin
   big_spec    = float(big_spec)
   big_err     = float(big_err)
endelse
writefits,root+'_wspec.fits',big_wave,hout
writefits,root+'_spec.fits',big_spec,hout
writefits,root+'_err.fits',big_err,hout

; Output object profile and wavelength calibration information.
openw,profile_lun,root+'_profile.dat',/get_lun
for i=0,n_orders-1 do $
   printf,profile_lun,ordernum(i),medwave_list(i),fwhm_list(i),$
          position_list(i),offset_list(i),$
          format='(f8.2,f14.3,3(f12.4))'
free_lun,profile_lun

; Output trace information
openw,trace_lun,root+'_trace.dat',/get_lun
for i=0,n_orders-1 do printf,trace_lun,ordernum(i),trace_coeffs(*,i),$
                             format='(6(f13.6))'
free_lun,trace_lun

; Final fit to object profile and trace parameters (for evaluating
; the fits post-processing)
indep_ls = where(independent_fit_made eq 1,n_indep,$
                 complement=fitted_ls,ncomp=n_fitted)
if (n_indep ge 1) then begin
   ; Subtract high-order terms
   net_fwhm   = fwhm_list         - fwhm_hot
   net_offset = offset_list       - offset_hot
   net_slope  = trace_coeffs(1,*) - slope_hot
   ; Fit residuals
   if (n_indep ge n_for_linear) then begin
      pfit_fwhm   = robust_poly_fit(ordernum(indep_ls),$
                                    net_fwhm(indep_ls),$
                                    fwhm_order)
      pfit_offset = robust_poly_fit(ordernum(indep_ls),$
                                    net_offset(indep_ls),$
                                    offset_order)
      pfit_slope  = robust_poly_fit(ordernum(indep_ls),$
                                    net_slope(indep_ls),$
                                    slope_order)
   endif else begin
      pfit_fwhm   = poly_fit(ordernum(indep_ls),net_fwhm(indep_ls),$
                             fwhm_order)
      pfit_offset = poly_fit(ordernum(indep_ls),net_offset(indep_ls),$
                             offset_order)
      pfit_slope  = poly_fit(ordernum(indep_ls),net_slope(indep_ls),$
                             slope_order)  
   endelse
   ; Reconstruct total fitted values
   fwhm_fit   = fwhm_hot
   offset_fit = offset_hot
   slope_fit  = slope_hot
   for k=0,fwhm_order do $
      fwhm_fit = fwhm_fit + pfit_fwhm(k)*(ordernum^k)
   for k=0,offset_order do $
      offset_fit = offset_fit + pfit_offset(k)*(ordernum^k)
   for k=0,slope_order do $
      slope_fit = slope_fit + pfit_slope(k)*(ordernum^k)
endif else begin
   fwhm_fit   = 0.*ordernum
   offset_fit = 0.*ordernum
   slope_fit  = 0.*ordernum
endelse
; Reconstruct fitted position
position_fit = ordercenter + offset_fit

; Plot profile and trace information
fwhm_plotmin   = min(fwhm_list) < min(fwhm_fit)
fwhm_plotmax   = max(fwhm_list) > max(fwhm_fit)
offset_plotmin = min(offset_list) < min(offset_fit)
offset_plotmax = max(offset_list) > max(offset_fit)
slope_plotmin  = min(trace_coeffs(1,*)) < min(slope_fit)
slope_plotmax  = max(trace_coeffs(1,*)) > max(slope_fit)
ps_openfile,root+'_trace',/portrait
   !p.multi    = [0,2,3]
   !p.charsize = 1.2
   !x.charsize = 1.2
   !y.charsize = 1.2
   !x.range    = [min(ordernum)-0.5,max(ordernum)+0.5]
   !x.style    = 1
   ; FWHM
   plot,ordernum,fwhm_list,psym=1,title='FWHM',$
        yr=[fwhm_plotmin,fwhm_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),fwhm_list(fitted_ls),psym=6
   oplot,ordernum,fwhm_fit
   plot,ordernum,fwhm_list-fwhm_fit,psym=1,title='FWHM Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(fwhm_list-fwhm_fit)(fitted_ls),psym=6
   ; Offset
   plot,ordernum,offset_list,psym=1,title='Offset',$
        yr=[offset_plotmin,offset_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),offset_list(fitted_ls),psym=6
   oplot,ordernum,offset_fit
   plot,ordernum,offset_list-offset_fit,psym=1,title='Offset Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(offset_list-offset_fit)(fitted_ls),psym=6
   ; Slope
   plot,ordernum,trace_coeffs(1,*),psym=1,title='Slope',$
        yr=[slope_plotmin,slope_plotmax]
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),trace_coeffs(1,fitted_ls),psym=6
   oplot,ordernum,slope_fit
   plot,ordernum,trace_coeffs(1,*)-slope_fit,psym=1,title='Slope Residuals'
   if (n_fitted ne 0) then $
      oplot,ordernum(fitted_ls),(trace_coeffs(1,*)-slope_fit)(fitted_ls),psym=6
ps_closefile

;; Remove temporary files.
print,'Removing temporary files...'
spawn,'"rm" temp_mage_err.fits'
spawn,'"rm" temp_mage_profile.ps'
spawn,'"rm" temp_mage_sky.fits'
spawn,'"rm" temp_mage_spec.fits'
spawn,'"rm" temp_mage_spec.ps'
spawn,'"rm" temp_mage_sub.fits'
spawn,'"rm" temp_mage_trace.ps'
spawn,'"rm" temp_mage_trace.dat'
spawn,'"rm" temp_mage_var.fits'
spawn,'"rm" temp_mage_wav.fits'

print,'Done.'

; Close log
journal

; Restore plot settings
cleanplot,/silent

return
end
