Pro mage_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile,$
                  flat2=flat2

;+--------------------------------------------------------------------------
;
; MAGE_SLITGRID      1/2008
;
; Trace orders in a bright standard star exposure to create a map of slit
; positions.  The slit position at a pixel is the x-intercept of the
; curve of constant position along the slit that would pass through
; the center of that pixel.
;
; Standard star should be at low airmass (less than 1.1)
;
; Assumes illuminated region is 1024x2048 (1x1 binning)
;
; NOTE: Rather than identyfying the order edges by hand, it may be
;       easier to determine the order positions using the known size
;       of the slit (assuming the projected size doesn't change much
;       across the array).
;
; INPUTS
;         stdfile   - String, FITS filename of standard star exposure
;
; KEYWORDS
;         flatfile  - String, FITS filename of flatfield exposure
;                     (non-diffused).  Used to identify the edges
;                     of each order.
;         flat2     - String, FITS filename of second flatfield
;                     exposure to use to identify order edges.
;                     If 'flatfield' is an internal lamp, then this
;                     should be a dome flat, and visa versa.  The
;                     two flats will be added.
;
; OUTPUTS
;         slitgrid  - Name of FITS file to which to write slit grid
;         orderlist - Name of ASCII file to which to write list of
;                     order centers and edges.  Output only if the
;                     flatfile keyword is set.
;
; HISTORY
;         Adapted from MIKE_SLITGRID 1/20/2008
;         Modified way in which order centers are calculated 9/16/2011
;---------------------------------------------------------------------------

if (n_params() lt 2) then begin
 print,'CALLING SEQUENCE: '
 print,'   mage_slitgrid,stdfile,slitgrid,orderlist,flatfile=flatfile'
 return
endif

device,decomp=0
loadct,39

;;; Overscan sizes
oscan_xsz = 128
oscan_ysz = 128

;;; Rough size of slit, including a buffer
slitsz = 46  ; 1x1 binning

;;; Read in standard star exposure
std  = mage_subbias(readfits(stdfile,hstd))
sz   = size(std)
xsz  = sz(1)
ysz  = sz(2)
illum_xsz = xsz - oscan_xsz
illum_ysz = ysz - oscan_ysz

;;; Arrays of x and y values
xarr = fltarr(xsz,ysz)
yarr = fltarr(xsz,ysz)
for i=0,xsz-1 do xarr(i,*) = i
for i=0,ysz-1 do yarr(*,i) = i

;;; Median cut through the array
cut_xmin = xsz/2.
cut_xmax = cut_xmin + 20
cut_xmid = 0.5*(cut_xmin+cut_xmax)
cut      = median(std(cut_xmin:cut_xmax,*),dim=1)

;;; Locate peaks.  Avoid edges of illuminated region.
first_peak = 1
med_cut    = median(cut)
cut_ymin   = 10
cut_ymax   = illum_ysz-11
for j=cut_ymin,cut_ymax do begin
   med_ymin = (0 > (j-20)) < (illum_ysz-1-40)
   med_ymax = med_ymin + 40
   if (cut(j) eq max(cut(j-10:j+10)) and $
       cut(j) gt median(cut(med_ymin:med_ymax)) + $
                    7*robust_sigma(cut(med_ymin:med_ymax)) and $
       cut(j) lt 60000. and $          ; Avoid saturated lines
       cut(j) ne cut(j-1)) then begin  ; Avoid double-counting
      if (first_peak) then begin
         peak_list  = j
         first_peak = 0
      endif else begin
         peak_list  = [peak_list,j]
      endelse
   endif
endfor
n_peak = n_elements(peak_list)

;;; Plot peak locations
window,9,xs=1000,ys=500,title='Peak Locations'
!p.multi = [0,1,1]
plot,cut,psym=10
for i=0,n_peak-1 do oplot,replicate(peak_list(i),2),[-1e6,1e6],color=250

;;; Illuminated regions
xrange = [0,illum_xsz-1]
yrange = [0,illum_ysz-1]

;;; Master slit grid.  Flag non-illuminated regions.
slit = fltarr(xsz,ysz) - 1000

;;; Y-intercept of each trace
y_intercept = fltarr(n_peak)

;;; Trace, fit, and compute slit grid in the vicinity of each peak.
print,'Tracing...'
window,10,xs=1000,ys=800,title='Traces'
!p.multi = [0,ceil(sqrt(n_peak)),ceil(sqrt(n_peak))]
for j=0,n_peak-1 do begin
   ; Avoid under-illuminated corners
   xstart = cut_xmid
   ystart = peak_list(j)
   temp_xmin = (400 - (400./500.)*ystart) > 0      ; 1x1 binning  
   temp_xmax = (1400 + (647./700.)*ystart) < 2047  ;
   traceline,std,xstart,ystart,xtr,ytr,/xdir,merit=merit,$
             median=3,range=xrange,bound=yrange,nside=10,/noplot,/bygauss
   ; Set trace fit polynomial.  Use a lower-ordr fit for the
   ; bluest orders.
   xlength = max(xtr)-min(xtr)
   if (xlength lt 1500) then npoly_trace = 3
   if (xlength ge 1500 and xlength lt 1800) then npoly_trace = 4
   if (xlength gt 1800) then npoly_trace = 5
   temp_trcoeffs = robust_poly_fit(xtr,ytr,npoly_trace,yfit,yerr)
   ; Reject outliers and refit to get uncertainties in
   ; coefficients.
   ls = where(abs(ytr-yfit) le 3*yerr)
   trcoeffs = poly_fit(xtr(ls),ytr(ls),npoly_trace,$
                       sigma=sig,yerror=yerr,yfit=yfit)
   y_intercept(j) = trcoeffs(0)
   ; Plot residuals
   plot,xtr(ls),ytr(ls)-yfit,psym=3,xs=3,/ystyle,yr=[-1,1],$
        title=strtrim(j,2)
   oplot,[-1e6,1e6],[0,0],linestyle=2
   ; Create a temporary 2D slit array assuming all loci of constant
   ; slit position run parallel to this trace
   temp_slit = yarr
   for k=1,npoly_trace do temp_slit = temp_slit - trcoeffs(k)*(xarr^k)

   ; Extract the region around this trace and save in the master
   ; slitgrid.  Assume the std is at the center of the slit, and
   ; include a margin at the edge of the slit.  Use wider margins for
   ; bluer orders.  Only save the blue end of the reddest order.
   if (j eq n_peak-1) then separation = (peak_list(j)-peak_list(j-1)) $
      else separation = (peak_list(j+1)-peak_list(j))
   if (y_intercept(j) le 862) then begin
     order_ls = where(temp_slit ge trcoeffs(0) - separation/2. and $
                      temp_slit le trcoeffs(0) + separation/2.)
   endif else begin
     order_ls = where(temp_slit ge trcoeffs(0) - separation/2. and $
                      temp_slit le trcoeffs(0) + separation/2. and $
                      xarr ge 940)
   endelse
   slit(order_ls) = temp_slit(order_ls)
endfor

;;; Write slitgrid to file
sxaddpar,hstd,'BZERO',0.0
writefits,slitgrid,slit,hstd

;;; Identify order edges (optional)
if keyword_set(flatfile) then begin
   flat = mage_subbias(readfits(flatfile,hflat))
   ; Use additional flat field?
   if keyword_set(flat2) then begin
      extra_flat = mage_subbias(readfits(flat2))
      flat       = flat + extra_flat
   endif
   window,11,xs=1000,ys=500
   !p.multi   = [0,1,1]
   print,'Left-click once on each edge of each order, left then right.'
   print,'Right-click when done with each order.'
   slit_slice = slit(300:1700,0:illum_ysz-1)
   flat_slice = flat(300:1700,0:illum_ysz-1)
   orderedge_list = fltarr(2*n_peak)
   for i=0,n_peak-1 do begin
      x_min = y_intercept(i) - (2./3)*slitsz
      x_max = y_intercept(i) + (2./3)*slitsz
      y_min = 0
      plot_ls = where(slit_slice ge x_min and slit_slice le x_max)
      y_max = cutvalue(flat_slice(plot_ls),99.9)
      plot,slit_slice(plot_ls),flat_slice(plot_ls),psym=3,xr=[x_min,x_max],/xs,$
           yr=[y_min,y_max]
      getpoints,temp_orderedge,/plotx,color=150
      orderedge_list(2*i:2*i+1) = temp_orderedge
   endfor
   
   ;;; Compute center of each order
   n_orderedge = n_elements(orderedge_list)
   if (n_orderedge mod 2 ne 0) then begin
      print,'MAGE_SLITGRID: Error - There should be an even number of order' 
      print,'               edges marked (two for each order).  Exiting without'
      print,'               writing order list.'
      return
   endif
   n_order        = n_orderedge / 2
   leftedge_list  = orderedge_list(2*lindgen(n_order))
   rightedge_list = orderedge_list(1+2*lindgen(n_order))
   ;center_list    = (leftedge_list + rightedge_list) / 2.
   ; Fit a smooth function to the order midpoints
   midpoint_list  = (leftedge_list + rightedge_list) / 2.
   center_polyorder = floor(n_order/2.) < 5
   ordernum         = findgen(n_order)
   center_pfit      = robust_poly_fit(ordernum,midpoint_list,center_polyorder)
   center_list      = 0.*midpoint_list
   for i=0,center_polyorder do center_list = center_list + $
                                             center_pfit(i)*(ordernum^i)

   ;;; Write order info to file
   print,'Writing order info to file...'
   openw,unit,orderlist,/get_lun
   printf,unit,'#  CCD   Order    Center   Left edge  Right edge'
   for i=0,n_order-1 do $
      printf,unit,'MAGE',i+1,center_list(i),leftedge_list(i),$
                  rightedge_list(i),format='(a6,i6,3(f12.3))'
   free_lun,unit
endif

; Restore plot settings
cleanplot,/silent

return
end
